/**************************************************************************\
*									   *
*   LNK1.C   -	  unido	X-Mac, Linker, Teil 1 von 3			   *
*									   *
*					Michael	Roehner, unido ChiPSy	   *
*									   *
\**************************************************************************/



/*
 * Aufruf :
 *
 * LNK [-c#] [-d#] [-l]	[-ofname] prog [module/library ... ]
 *
 * oder	:
 *
 * LNK [-c#] [-d#] [-l]	[-ofname] cmdfile
 *
 *
 *   -c#		Absolutadresse des Programmes ist # (hex);
 *			Ergebnis heisst	"prog.BIN" und nicht
 *			"prog.COM". Bei	dieser Option wird ein
 *			Sprung zum Programmstart nicht am Anfang
 *			eingefuegt (nur	bei COM).
 *
 *   -d#		Absolutadresse der Daten ist # (hex).
 *			Es wird	ein extra Datenfile angelegt,
 *			"prog.DAT". Wenn diese Option nicht
 *			angegeben wird,	wird das Datensegment hinter
 *			das Codesegment	gelinkt	(dann 2-Pass-Link, sonst
 *			bei -D 1-Pass-Link).
 *
 *   -l			Aktivitaeten des Linkers ausgeben.
 *
 *   -ofname		Angabe des Outputfilenamens fname, Extension
 *			muss mitangegeben werden. Wenn diese Option
 *			nicht angegeben	wird, dann ist prog der
 *			Outputfilename (+ Extension).
 *
 *   prog		Dateiname des zu linkenden Programms.
 *			Als Extension ist nur REL zugelassen.
 *
 *   module/library...	Eine Liste mit Modulen (.REL) und/oder
 *			Bibliotheken (.LIB). Alle Module werden	zusammen-
 *			gelinkt	und die	angegebenen Bibliotheken werden
 *			nach noch nicht	aufgeloesten externen Referenzen
 *			durchsucht.
 *
 *   cmdfile		Dies ist eine Textdatei	(.LNK),	in der alle obigen
 *			Optionen und Dateinamen	stehen koennen,	jede Zeile
 *			eine Option oder ein Dateiname.	Jede Zeile wirkt wie
 *			ein Argument in	der Aufrufzeile. Zusaetzlich
 *			zu den Argumenten in dieser Datei koennen weitere
 *			Optionen oder Namen auch in der	Aufrufzeile
 *			angegeben werden.
 *
 *
 *   Die Deklaration eines externen Symbols reicht schon aus,
 *   damit das betreffende Modul geladen wird. Es muss im Programm
 *   gar nicht erst verwendet werden !
 *
 *   Dateinamen	ohne Extension oder mit	Extension REL identifizieren Module;
 *   die Extension .LIB	identifiziert Bibliotheken. Falls noetig, erfolgt
 *   ein erneuter Durchgang durch eine Bibliothek, um Rueckwaertsverweise in
 *   Bibliotheksmodulen	aufzuloesen. Module und	Bibliotheken werden in der
 *   in	der Kommandozeile angegebenen Reihenfolge verarbeitet.
 *
 *
 *   Datei Extensions :
 *
 *	-  bei Bibliotheken muss .LIB angegeben	werden
 *
 *	- .rel = relokatibles Objektmodul
 *	  .lib = Bibliothek mit	Objektmodulen
 *	  .com = Transientes Kommando von CP/M (Linker-Output)
 *	  .bin = Ergebnis von -C#
 *	  .dat = Daten in extra	Datei, wenn -D#
 *	  .lnk = Linkkommandodatei
 *
 *
 * -------------------------------------------------------------------------
 *
 *  16-Bit-Eintraege werden immer in High-Byte-	/ Low-Byte-Reihenfolge
 *  abgelegt !
 *
 *
 *
 *   Codebuffer	:		     cbuffer ---> +----------+	  <---
 *						  |	     |	    |
 *						  |	     |
 *   cbuffer			      cbnext ---> |	     |	  CDBSZ
 *     zeigt auf den Anfang		      |	  |	     |	  Bytes
 *     des Codesegments			      v	  |	     |
 *						  |	     |	    |
 *   cbnext
 *     zeigt auf das naechste			       .	    .
 *     Byte im Codesegment			       .	    .
 *
 *   dbuffer / dbnext				  |	     |	    |
 *     s.o. fuer Datensegment,		      A	  |	     |	    |
 *     das Datensegment	waechst		      |	  |	     |	    |
 *     von oben	nach unten	      dbnext ---> |	     |	    |
 *     dem Codesegment				  |	     |	    |
 *     entgegen			     dbuffer ---> +----------+	  <---
 *
 *
 *  16-Bit-Indizes im  Codesegment  u n	d  im Datensegment werden High-Byte /
 *  Low-Byte gespeichert. Indizes zeigen dabei im Datensegment auf das
 *  Low-Byte.
 *
 *
 *
 *   Tabelle relativer Indizes und Symboltabelle :
 *
 *
 *				    +-----+---------+			<---
 *			crstart	--> | Typ | Wert    |			  |
 *				    +-----+---------+
 *				    |	  |	    |			STBSZ
 *				    +-----+---------+			Bytes
 *				    |	  |	    |
 *			 crnext	-->					  |
 *					  .				  .
 *					  .				  .
 *
 *				    |	  |	    |			  |
 *				    +-----+---------+			  |
 *				    |	  |	    |			  |
 *				    +-----+---------+			  |
 *			  crend	-->
 *					  .				  .
 *					  .				  .
 *
 *			   +---------+---------+--------------+-----+	  |
 *     snext -->	   |	     |	       |	      |	    |	  |
 *			   +---------+---------+--------------+-----+	  |
 *		       +-> |	     |	       |	      |	    |	  |
 *		       |   +---------+---------+--------------+-----+	  |
 *
 *					  .				  .
 *		     v	 A		  .				  .
 *		     |	 |
 *	       v   A |	 | +---------+---------+--------------+-----+	  |
 *	       |   | |	 | |	     |	       |	      |	    |	  |
 *	       |   | |	 | +---------+---------+--------------+-----+	  |
 *	       |   | |	 | |	     |	       |	      |	    |	  |
 *	       |   | |	 | +---------+---------+--------------+-----+	  |
 *	       |   | |+->+-|	     |	       |	      |	    |	  |
 *	       |   | ||	   +---------+---------+--------------+-----+	  |
 *	       |+->+-------|	     |	       |	      |	    |	  |
 *	       ||    ||	   +---------+---------+--------------+-----+	  |
 *	       ||    +---->|	     |	       | HIGH,0	      |	    |	  |
 *	       ||     |	   +---------+---------+--------------+-----+	  |
 * xrfirst --> ||     +--- |	     |	       | ""	      |	    |	  |
 *	       ||	   +---------+---------+--------------+-----+	  |
 *	       +---------->|	     |	       | HIGH,0	      |	    |	  |
 *		|	   +---------+---------+--------------+-----+	  |
 * epfirst -->	+----------|	     |	       | ""	      |	    |	  |
 *			   +---------+---------+--------------+-----+	<---
 *			     NXT       VAL	 SYM	       TYP
 *
 *
 *
 *   crstart   -  Zeiger auf Anfang Indizes relative Eintraege
 *		  (auch	Basis fuer Symboltabellenindizes)
 *
 *   crnext    -  Zeiger auf naechstes Byte Indizes relative Eintrage
 *
 *   crend     -  Zeiger hinter	letztes	Byte Indizes relative Eintraege
 *		  (nur waehrend	Phase 2)
 *
 *   Wert-Feld -  entweder Offset fuer naechste	Referenz oder Index
 *		  in Code- oder	Datensegment (ab .buffer) auf relativen
 *		  Eintrag, 2 Byte
 *
 *   Typ-Feld  -  Typ des Wertefelds an, 1 Byte
 *
 *		    7 6	5 4 3 2	1 0
 *		    |	    ---	---
 *		    |	     |	 |
 *		    |	     |	 +---> ABS/PREL/DREL, Typ des Wertes
 *		    |	     |
 *		    |	     +-------> CSEGBIT/DSEGBIT,	Typ des	Segments
 *		    |
 *		    +----------------> XRPBIT, 1 = Plusoffset (andere Bits 0)
 *
 *
 *
 *   NXT-Feld  -  Index	auf naechsten Eintrag in Kette,	2 Byte
 *		    0 =	Ende einer Liste
 *		    (Zeiger = crstart +	getuint(sptr[NEXT]))
 *
 *   VAL-Feld  -  Wert des Symbols, 2 Byte
 *
 *   SYM-Feld  -  Symbolstring nullterminiert, MAXSYM+1	Bytes
 *		    stptr[SYM] == '\0' Listenanfang
 *		    stptr[SYM] == HIGH Listenende
 *
 *   TYP-Feld  -  Typ des Symbols, 1 Byte
 *		    PREL  -  programmrelativ
 *		    DREL  -  datenrelativ
 *		    ABS	  -  absolut
 *
 *   snext     -  Zeiger auf ersten freien Eintrag in Symboltabelle
 *		  (Symboltabelle waechste nach unten)
 *
 *   xrfirst   -  Zeiger auf verkettete	Liste der Externals
 *
 *   epfirst   -  Zeiger auf verkettete	Liste der Einsprungpunkte
 *
 *
 *   Die Listen, auf die epfirst/xrfirst zeigen, sind alphabetisch sortiert.
 *   Sie fangen	an (erstes Symbol) jeweils mit dem alphabetisch	"kleinsten"
 *   Leerstring	"" und werden beendet durch den	alphabetisch "hoechsten"
 *   String HIGH,0. Nach dem Initialisieren der	Listen bestehen	die Listen
 *   ausschliesslich aus jeweils diesen	beiden Elementen, dies vereinfacht
 *   das Einfuegen der "eigentlichen" Symbole (keine Sonderfaelle beim
 *   Listenanfang- und Ende sowie ersten Eintrag).
 *
 *
 */





/*							       Includes	  */
/**************************************************************************/

#include <stdio.h>			/* Standardausgabe benutzt	  */
#include <string.h>			/* Stringfunktionen benutzt	  */
#include <stdlib.h>
#include <ctype.h>
#include "all.h"			/* allgemeine Festlegungen	  */
#include "lnk.h"			/* Definitionen	fuer Linker	  */
#include "lnke.h"			/* Externals LNK.C		  */
#include "rel.h"			/* Aufbau REL			  */
#include "uti.h"
#include "rele.h"			/* Externals REL.C		  */

/*------------------------------------------------------------------------*/
/*--- Deklaration lokaler Funktionen				       ---*/
/*------------------------------------------------------------------------*/
void usage __ARGS((void));
void newtbl __ARGS((char **));

/**************************************************************************\
*									   *
* action      :	 Hauptprogramm.						   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 argc, argv  - Aufrufparameter				   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 argbuf	  -  Zeiger auf	Anfang Argumentenbuffer		   *
*		 myargv	  -  eigenes argv[]				   *
*		 myargc	  -  eigenes argc				   *
*		 argfn	  -  Name LNK-Datei, falls angegeben		   *
*		 argfile  -  LNK-Datei					   *
*		 monitor  -  TRUE, wenn	Option -L bei Aufruf angegeben	   *
*		 bin	  -  TRUE, wenn	Option -C bei Aufruf angegeben	   *
*		 cbase	  -  Wert der Option C,	sonst COMBASE		   *
*		 dbase	  -  Wert der Option D,	sonst 0			   *
*		 data	  -  YES wenn Option D angegeben		   *
*		 cbuffer  -  Zeiger auf	Anfang Codebufferspeicher	   *
*		 cnext	  -  = cbuffer					   *
*		 dbuffer  -  Zeiger auf	Anfang Datenbufferspeicher	   *
*		 dnext	  -  = dbuffer (waechst	nach unten)		   *
*		 crstart  -  Zeiger auf	Anfang External-Codewerte-Buffer   *
*		 crnext	  -  = crnext					   *
*		 snext	  -  erster freie Symbolbuffer (von oben)	   *
*		 epfirst  -  Zeiger auf	initialisierte Entrypoint-Liste	   *
*		 xrfirst  -  Zeiger auf	initialisierte Externals-Liste	   *
*		 outfn	  -  Outputdateiname wenn Option -O beim Aufruf	   *
*									   *
* locals      :	 s.u.							   *
*									   *
* returns     :	 Errorcode 1 bei Fehlern				   *
*		 Errorcode 0 sonst					   *
*									   *
\**************************************************************************/

void main(argc,argv)

int	argc;
char   *argv[];

  {
    char   arg[FNSIZE+1];     /* Buffer	fuer aktuelle Option aus Aufruf	  */
    int	   argnbr;	      /* Nummer	der aktuellen Option in	arg	  */
    int	   b;		      /* Zwischenspeicher fuer Zahl in Option	  */
    int	   len;		      /* Laenge	Zahl in	Option			  */
    char  *cptr;	      /* fuer das Scannen nach LNK-Datei	  */



    /* Startmeldung (Identifikation) ausgeben */

    puts("NORD><LINK Z80 REL-Linker Version 1.1");
    puts("ported to Linux by dl4sdc");


    /* Speicher	fuer holen und zuweisen	(nicht dynamisch) */

    if (!(argbuf = (char *)malloc(ARGBSZ)))		   /* zuweisen	  */
      fatal("Not enough	buffer memory","");		   /* kein Platz  */
    if (!(crnext = crstart = (char *)malloc(STBSZ)))	   /* zuweisen	  */
      fatal("Not enough	table memory","");		   /* kein Platz  */
    if (!(cbuffer = cbnext = (char *)malloc(CDBSZ)))	   /* zuweisen	  */
      fatal("Not enough	memory for code","");		   /* kein Platz  */
    dbuffer = dbnext = cbuffer + (long)(CDBSZ -	1);
    snext = crstart + (long)(STBSZ - SSZ);	  /* 1.	freie E. */
    newtbl(&epfirst);				  /* Init Entrypointliste */
    newtbl(&xrfirst);				  /* Init External-Liste  */



    /* "mein" argv erstellen, LNK-Datei	abarbeiten */

    myargc = argnbr = 0;			       /* Start		  */
    cptr = argbuf;				       /* Argumentbuffer  */
    while (argnbr < argc)			       /* alle Argumente  */
      {						       /* durchgehen	  */
	/* strupper(argv[argnbr]);			  alles	gross	  */
	if (	*argv[argnbr] != '-'		       /* name.LNK ?	  */
	     &&	(len = strlen(argv[argnbr])) > 4
	     &&	!strcmp(LNKEXT,argv[argnbr] + len - 4)
	   )
	  {					       /* LNK-Datei	  */
	    strcpy(argfn,argv[argnbr++]);	       /* oeffnen	  */
	    if (!(argfile = fopen(argfn,"rt")))
	      fatal(argfn," - Cannot open");
	    while (fgets(cptr,FNSIZE+4,argfile))       /* einlesen	  */
	      if (    (len = strlen(cptr)) > 1
		   && *cptr != '\n'
		   && *cptr != ' '
		 )
		{
		  if (	  myargc == ARGMAX
		       || cptr + 2 * (FNSIZE + 4) >= argbuf + ARGBSZ
		     ) fatal ("Argument	buffer overflow","");
		  myargv[myargc++] = cptr;	       /* Args.	zuordnen  */
		  /* strupper(cptr); */
		  cptr += len -	1;
		  if (*cptr != '\n') ++cptr;
		  *cptr++ = '\0';
		}
	    continue;				       /* naechstes Arg.  */
	  }
	myargv[myargc++] = argv[argnbr++];	       /* Arg. kopieren	  */
      }



    /* Schalter	aus Kommandozeile holen	und merken */

    cbase = COMBASE;
    argnbr = 0;
    while (getarg(++argnbr,arg,FNSIZE+3,myargc,myargv) != EOF)
      {
	if (arg[0] != '-') continue;	     /*	Dateinamen uebergehen	  */
	switch (arg[1])
	  {
	    case 'c' :	bin = YES;		       /* -c<hexadresse>  */
			len = xtoi(arg + 2,&b);
			if (len	>= 0 &&	!arg[len + 2])
			  cbase	= b;
			else
			  usage();
			break;

	    case 'd' :	data = YES;		       /* -d<hexadresse>  */
			len = xtoi(arg + 2,&b);
			if (len	>= 0 &&	!arg[len + 2])
			  dbase	= b;
			else
			  usage();
			break;

	    case 'l' :	monitor	= YES;		       /* -l		  */
			break;

	    case 'o' :	if (strlen(arg)	> 2)	       /* -o<filename>	  */
			  strcpy(outfn,arg + 2);
			else
			  usage();
			break;

	    default  :	usage();		       /* Optionsfehler	  */
			break;
	  }
     }



    /* Objektcode laden	und linken */

    phase1(myargc,myargv);
    if (!okay()) exit(1);		     /*	Unaufgeloestes -> Abbruch */
    if (!data)				     /*	2. Durchlauf noetig ?	  */
      {
	puts("\n      second pass");	     /*	2. Durchlauf ...	  */
	if (start && !bin) cbase += 3;	     /*	ggf. JP	zum Start noetig  */
	dbase =	cbase +	csize;		     /*	Datenbasis festlegen	  */
	sfree =	cloc = dloc = 0;	     /*	Variable initialisieren	  */
	crnext = crstart;
	cbnext = cbuffer;
	dbnext = dbuffer;
	snext =	crstart	+ (long)(STBSZ - SSZ);
	newtbl(&epfirst);
	newtbl(&xrfirst);
	phase1(myargc,myargv);		     /*	2. Durchlauf durchfuehren */
      }



    /* Objektcode ausgeben  */

    phase2();				     /*	Code schreiben		  */
    exit(0);
    /* return(FALSE); */
  }





/**************************************************************************\
*									   *
* action      :	 Wenn Zeichen c	ein Kleinbuchstabe ist,	umwandeln in	   *
*		 Grossbuchstabe.					   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 c  -  umzuwandelndes Zeichen				   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 c umgewandelt (s.o.)					   *
*									   *
\**************************************************************************/

int chupper(c)

int c;

  {
    return ((c>='a' && c<='z') ? c - ('a' - 'A') : c);
  }





/**************************************************************************\
*									   *
* action      :	 Alle Kleinbuchstaben im String	str in Grossbuchstaben	   *
*		 umwandeln.						   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 str  -	 Zeiger	auf nullterminierten String		   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

void strupper(str)

char *str;

  {
    while(*str)
      {
	*str = chupper(*str);
	++str;
      }
    /* return(FALSE); */
  }





/**************************************************************************\
*									   *
* action      :	 n. Argument nach s aus	Argumentenstringarray argv	   *
*		 (enthaelt argc	Argumentstrings) holen.			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 n	-   Nummer des Arguments (0 -> *argv[0])	   *
*		 s	-   Zeiger auf den Zielstring (= *arv[n])	   *
*		 size	-   maximale Anzahl Zeichen zu kopieren		   *
*		 argc	-   Anzahl Argumente an	*argv			   *
*		 argv	-   Zeiger auf Array von Strings, jeder	ist ein	   *
*			    Argument					   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 str	-  Zeiger in Argument in *argv			   *
*		 i	-  Zeichenzaehler Argumentstring		   *
*									   *
* returns     :	 Anzahl	der kopierten Zeichen oder EOF,	wenn nichts	   *
*		 gefunden.						   *
*									   *
\**************************************************************************/

int getarg(n,s,size,argc,argv)

int	n;
char   *s;
int	size;
int	argc;
char   *argv[];

  {
    char   *str;
    int	    i;

    if (n < 0 || n >= argc)		/* wenn	angefordertes Argument	  */
      {					/* nicht vorhanden, dann	  */
	*s = '\0';			/* leeren String zurueckgeben	  */
	return (EOF);			/* und Flag "nicht vorhanden"	  */
      }
    i =	0;				/* Anzahl kopierte Zeichen	  */
    str	= argv[n];			/* Zeiger auf n. Argumentstring	  */
    while (i < size)			/* Argumentstring kopieren,	  */
      {					/* dabei Zeichen zaehlen und nur  */
	if ((s[i] = str[i]) == '\0')	/* maximal size	- 1 Zeichen	  */
	  break;			/* (size - 1 weil bei Abbruch	  */
	++i;				/* wegen size noch Stringende	  */
      }					/* '\0'	dazukommt)		  */
    s[i] = '\0';			/* falls wegen size abgebrochen	  */
    return (i);				/* Kopierlaenge	zurueck		  */
  }





/**************************************************************************\
*									   *
* action      :	 Programmabbruch mit Bedienungshinweis.			   *
*									   *
\**************************************************************************/

void usage()
  {
    fatal("Usage: LNK [-c#] [-d#] [-l] [-ofname] prog [module/library ... ]\n",
	  "   or: LNK [-c#] [-d#] [-l] [-ofname] cmdfile");
  }





/**************************************************************************\
*									   *
* action      :	 Alphabetische Liste initialisieren.			   *
*									   *
*		 Erstes	Symbol setzen mit dem alphabetisch "kleinsten"	   *
*		 Leerstring "",	naechstes Symbol (1. letztes) setzen	   *
*		 mit dem alphabetisch "hoechsten" String HIGH,0.	   *
*		 Beide Symbole sind Dummysymbole, die das Einfuegen der	   *
*		 "eigentlichen"	Symbole	vereinfachen, da es keine	   *
*		 Sonderfaelle beim Listenanfang- und Ende sowie	ersten	   *
*		 Eintrag zu beachten gilt.				   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 low	-  Zeiger auf den zu setzenden Zeiger auf den	   *
*			   Listenanfang, wird gesetzt			   *
*									   *
* r/o globals :	 high	-  Zeiger auf String, der HIGH,0 enthaelt	   *
*									   *
* r/w globals :	 snext	-  Zeiger auf den naechsten freien Eintrag in	   *
*			   Symboltabelle (von oben)			   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

void newtbl(low)

char **low;

  {
    *low = snext;				  /* Listenanfangszeiger  */
    strcpy(snext + SYM,"");			  /* 1.	Symbol setzen	  */
    putuint(snext + NXT,snext -	SSZ - crstart);	  /* Kettung auf 2.	  */
    snext -= SSZ;				  /* naechster Freiplatz  */
    strcpy(snext + SYM,high);			  /* 2.	Symbol setzen	  */
    putuint(snext + NXT,0);			  /* Listenende		  */
    snext -= SSZ;				  /* naechster Freiplatz  */
  }





/**************************************************************************\
*									   *
* action      :	 Neuen Eintrag in Symboltabelle	alphabetisch hinter prev   *
*		 speichern (Suche nach alphabetischem Platz ab prev).	   *
*		 Programmabbruch mit Meldung, wenn Symboltabelle voll.	   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 prev	 -  Zeiger auf Zeiger in Symboltabelle,	hinter	   *
*			    *prev beginnt die Suche nach alphabetisch	   *
*			    richtiger Stelle, an der das neue Symbol	   *
*			    eingehaengt	wird, wird nach	Setzen des neuen   *
*			    Symbols auf	dieses gesetzt			   *
*									   *
* r/o globals :	 symbol	 -  Symbolstring des neuen Symbols, alle Symbole   *
*			    einer Liste	kommen in alphabetischer Reihen-   *
*			    folge bei newsym() an			   *
*		 crstart -  Zeigerbasis	der Symboltabelle		   *
*		 crnext	 -  hier muss Symboltabelle spaetestens	enden	   *
*		 type	 -  Typ	des neu	zu speichernden	Symbols		   *
*		 cmod	 -  Codeprogrammzaehlerstand bei Modulanfang	   *
*		 dmod	 -  Datenprogrammzaehlerstand bei Modulanfang	   *
*									   *
* r/w globals :	 sfree	 -  Index Liste	wieder freier Stellen in Symbolt.  *
*		 snext	 -  Zeiger naechste freie Stelle in Symboltabelle  *
*		 field	 -  Wert des Symbols, wird mit cmod oder dmod	   *
*			    korrigiert,	wenn es	ein relatives Symbol ist   *
*									   *
* locals      :	 cp	 -  Zeiger auf naechsten zu untersuchenden Eintrag *
*		 new	 -  Zeiger auf zu besetzenden neuen Eintrag	   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

void newsym(prev)

char **prev;

  {
    char   *cp;
    char   *new;

    if (sfree)				     /*	wenn "alte" freie	  */
      sfree = getuint(new = crstart + (long)sfree);/* Eintraege	vorhanden, dann	*/
    else				     /*	davon einen nehmen	  */
      {					     /*	sonst gucken, ob Platz	  */
	new = snext;			     /*	naechster leere	Eintrag	  */
	if ((snext -= SSZ) < crnext)	     /*	Symboltabelle voll ?	  */
	  fatal("Symbol	memory overflow","");/*	... ja,	Programmabbruch	  */
      }
    cp = *prev;				     /*	Anfang Suche Einfuegepos. */
    while (strcmp(symbol,cp + SYM) >= 0)     /*	Position finden		  */
      {					     /*	( = Eintrag, dessen	  */
	*prev =	cp;			     /*	Symbolstring alphabetisch */
	cp = crstart + (long)getuint(cp	+ NXT);	   /* groesser ist -  vor dem	*/
      }					     /*	einzufuegen ist)	  */
    putuint(new,cp - crstart);		     /*	Index Neu -> Nachfolger	  */
    putuint(*prev,new -	crstart);	     /*	Index Vorgaenger -> Neu	  */
    *prev = new;			     /*	Neu aktueller Vorgaenger  */
    switch(type)			     /*	programm- oder daten-	  */
      {					     /*	relative Werte auf	  */
	case PREL :  field += cmod;  break;  /*	Gesamtsegmentanfaenge	  */
	case DREL :  field += dmod;  break;  /*	umrechnen		  */
      }
    putuint(new	+ VAL,field);		     /*	Wert eintragen		  */
    *(new + TYP) = type;		     /*	Typ eintragen		  */
    strcpy(new + SYM,symbol);		     /*	Symbolstring eintragen	  */
  }





/**************************************************************************\
*									   *
* action      :	 Namen des naechsten Moduls aus	Input-REL/LIB-Datei holen, *
*		 dabei Datei durchsuchen nach dem naechsten Modulnamen.	   *
*		 Programmabbruch mit Meldung, wenn REL/LIB-Datei	   *
*		 fehlerhaft ist.					   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 (siehe	getrel())					   *
*									   *
* r/w globals :	 (siehe	getrel())					   *
*		 modname  -  Namensstring des naechsten	Moduls,	wenn	   *
*			     Modulnamenseintrag	in REL/LIB-Datei gefunden  *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 YES - wenn Modulname gefunden				   *
*		 NO  - wenn REL-Datei zu Ende				   *
*		 Programmabbruch bei Fehlern				   *
*									   *
\**************************************************************************/

BOOLEAN	getname()
  {
    LOOP					       /* solange Modul-  */
      {						       /* name nicht	  */
	switch(getrel())			       /* gefunden,	  */
	  {					       /* REL-Item holen  */
	    case PNAME :  strcpy(modname,symbol);      /* Modulname	  */
			  return (YES);
	    case EFILE :  return (NO);		       /* Fileende	  */
	    case ERR   :  fatal(infn," - Corrupted");  /* REL-Fehler	  */
	  }
      }
  }



/* Ende	von LNK1.C */
