/**************************************************************************\
*									   *
*   MAC1.C   -	  unido	X-Mac, Teil 1 von 5				   *
*									   *
*									   *
*					Michael	Roehner, unido ChiPSy	   *
*   03-11-90  Urversion	von DC4OX					   *
*   05-11-90  exit(0) bei Ende ohne Fehler			DF2AU	   *
*	      Signon gendert						   *
*	      Funktionen (ausser VOID) geben Wert zurueck		   *
*   26-12-90  Fehler in	Pointer	Arithmetik beseitigt		DF2AU	   *
*									   *
*									   *
*    -	 Dateinamen und	Directory-Wege werden sowohl mit / als auch mit	\  *
*	 unterstuetzt, bei anderen Wegangaben als			   *
*									   *
*	   [<drive>:\ |	.\ | ..\]{<directory>\}<filename>[.<extension>]	   *
*									   *
*	   oder								   *
*									   *
*	   [<drive>:/ |	./ | ../]{<directory>/}<filename>[.<extension>]	   *
*									   *
*	 muessen die die Dateinamen betreffenden Programmteile angepasst   *
*	 werden.							   *
*									   *
\**************************************************************************/



/*
 *   Aufruf :
 *
 *   mac <Dateispezifikation> [-l]
 *
 *
 *   MAC assembliert die angegebene Quelldatei und schreibt das	Ergebnis
 *   in	eine Datei desselben Names, aber mit Extension .REL. Der Modulname
 *   in	der REL-Datei entspricht den ersten 6 Zeichen des Dateinamens.
 *   Die Extension des Quelltextes ist default "MAC", falls nichts
 *   anderes angegeben wurde.
 *
 *   Ist der Schalter -l angegeben, so wird ein	Listing	an stdout
 *   ausgegeben.
 *
 *
 *
 * -------------------------------------------------------------------------
 *
 *
 *
 * Aufbau der Symboltabelle :
 *
 *			  +--- st
 *			  |
 *			  V	      MAXLAB Bytes   INTSZ Bytes   1 Byte
 *			 +-------------------------+-------------+--------+
 *			 | 0			   |		 |	  |
 *			 +-------------------------+-------------+--------+
 *
 *			 +-------------------------+-------------+--------+
 *  +------------------->| Symbol (ASCII)	   | Wert	 | Flags  |
 *  |			 +-------------------------+-------------+--------+
 *  |
 *  |			 +-------------------------+-------------+--------+
 *  |			 | 0			   |		 |	  |
 *  |			 +-------------------------+-------------+--------+
 *  |
 *  |			 +-------------------------+-------------+--------+
 *  |	       stptr -+->| Symbol (ASCII)	   | Wert	 | Flags  |
 *  |	stp	      |	 +-------------------------+-------------+--------+
 *  |	 |	      |
 *  |	 |	      |	 +-------------------------+-------------+--------+
 *  |	 V	      |	 | 0			   |		 |	  |
 *  |	+----------+  |	 +-------------------------+-------------+--------+
 *  |	|	   |--+
 *  |	+----------+	 +-------------------------+-------------+--------+
 *  +---|	   |	 | 0			   |		 |	  |
 *	+----------+	 +-------------------------+-------------+--------+
 *	|	   |---+
 *	+----------+   |		 .
 *	|	   |   |
 *	     .	       |		 .
 *	     .
 *	     .		 +-------------------------+-------------+--------+
 *			 | 0			   |		 |	  |
 *			 +-------------------------+-------------+--------+
 *									   ^
 *									   |
 *								  stend	---+
 *
 *
 *   stn = Anzahl der besetzten	(| Symbol (ASCII) ... |) Symbolslots,
 *	   gleich Anzahl der Eintrage in *stp
 *
 *   Symbole werden uppercase abgespeichert und	bis MAXLAB mit 0-en
 *   aufgefuellt.
 *
 *   Zu	einem gegebenen	Symbol wird der	Hashwert berechnet, dieser gibt	den
 *   Slot fuer das entsprechende Symbol	an. Ist	der Slot schon mit einem
 *   anderen Symbol besetzt, wird der sequentiell naechste Slot	getestet,
 *   dies wird mit Wrap-Around in der Symboltabelle gemacht.
 *
 *
 *
 * Aufbau der Makrotabelle :
 *
 *  +--- mt
 *  |
 *  V
 * +--------+--------+---+---------------+---+--     -+----------------+---+
 * | <next> | <Name> | 0 | <erste Zeile> | 0 |	 ...  |	<letzte	Zeile> | 0 |
 * +--------+--------+---+---------------+---+--     -+----------------+---+
 *     |
 *  +--+ (Index)
 *  V
 * +--------+--------+---+---------------+---+--     -+----------------+---+
 * | <next> | <Name> | 0 | <erste Zeile> | 0 |	 ...  |	<letzte	Zeile> | 0 |
 * +--------+--------+---+---------------+---+--     -+----------------+---+
 *     |				      ^
 *  +--+ (Index)			      |
 *  V			     .		      +--- mtptr
 *			     .
 *			     .
 *
 *  +--- mtprev
 *  |
 *  V
 * +--------+--------+---+---------------+---+--     -+----------------+---+
 * | <next> | <Name> | 0 | <erste Zeile> | 0 |	 ...  |	<letzte	Zeile> | 0 |
 * +--------+--------+---+---------------+---+--     -+----------------+---+
 *     |
 *  +--+ (Index)		    mtnext ---+
 *  V					      V
 * +--------+--------+---+---------------+---+--     -+----------------+---+
 * |  0000  | <Name> | 0 | <erste Zeile> | 0 |	 ...  |	<letzte	Zeile> | 0 |
 * +--------+--------+---+---------------+---+--     -+----------------+---+
 *
 *			     .
 *			     .
 *			     .
 *
 *  .......................................... <Ende Makrobuffer>
 *		     ^
 *		     |<------ MAXLINE ------>|
 *		     |
 *		     +--- mtend
 *
 *
 *   <next> =
 *   Index auf den naechsten Makroeintrag oder 0, wenn es keinen
 *   naechsten Eintrag gibt. Zeiger auf	den naechsten Eintrag
 *   errechnet sich durch mt + <next>.
 *
 *   <Name> =
 *   Name des Makros (bis MAXLAB Zeichen lang).
 *
 *   <erste Zeile> ... <letzte Zeile> =
 *   Die Textzeilen des	Makros,	aufgenommen werden alle	Zeilen	h i n t	e r
 *   der MACRO-Zeile bis zur ENDM-Zeile	einschliesslich.
 *
 *
 * Pass	2, Makroparameterersetzung und Makrolabel :
 *
 * Bei jedem Makroaufruf wird zunaechst	die Parameterliste hinter dem
 * Aufruf ab mtnext in Makrotextbuffer abgelegt	(mtnext	bleibt,	dies ist
 * ein temporaeres Ablegen nur fuer das	Abarbeiten des Makros).	Die
 * Parameter werden ab mtnext einzeln als nullterminierte Strings abgelegt,
 * Strings als Parameter ("..",'..') ohne die Delimiter. In mpptr[] steht
 * fuer	jeden Parameter	ein Zeiger auf den nullterminierten Parameterstring
 * im Makrotextbuffer oder NULL, wenn kein Parameter angegeben wurde.
 * Alle	Eintraege in mlnbr[] werden geloescht.
 *
 * Nach	dem Aufruf werden alle Makrozeilen einzeln geholt und die Parameter
 * anstatt der Platzhalter textuell eingesetzt.
 *
 * Makrolabel werden in	reale Label umgewandelt. In mlnext steht die
 * laufende Nummer des naechsten moeglichen realen Labels. Beim	ersten
 * Auftreten eines Makrolabels wird diesem mlnext zugeordnet, mlnext wird
 * erhoeht. Die	Zuordnungen werden in mlnbr[] fuer jede	Makrolabelnummer
 * abgespeichert waehrend der Makroabarbeitung und bei jedem weiteren
 * Auftauchen einer Makrolabelnummer wird die entsprechende Nummer durch
 * den Eintrag in mlnbr[] ersetzt.
 *
 *
 */





/*							       Includes	  */
/**************************************************************************/

#include <stdio.h>			/* Standardausgabe benutzt	  */
#include <string.h>			/* Stringfunktionen benutzt	  */
#include <ctype.h>
#include <stdlib.h>
#ifdef MSDOS
#include <alloc.h>
#endif
#include "all.h"			/* allgemeine Festlegungen	  */
#include "mac.h"			/* Festlegungen	fuer MAC	  */
#include "mace.h"			/* Externals MAC		  */
#include "rel.h"			/* Aufbau REL			  */
#include "uti.h"
#include "rele.h"			/* Externals REL.C		  */

/* Prototypen der lokalen nicht-integer	Funktionen					  */
void pass1 __ARGS((void));
void pass2 __ARGS((void));
void dopass __ARGS((void));
void usagerr __ARGS((void));

/**************************************************************************\
*									   *
* action      :	 Hauptprogramm.						   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 argc, argv  - Aufrufparameter				   *
*									   *
* r/o globals :								   *
*									   *
* r/w globals :	 srcfn							   *
*		 objfn							   *
*		 st							   *
*		 stend							   *
*		 stp							   *
*		 mt							   *
*		 mtnext							   *
*		 mtend							   *
*		 srcfile						   *
*		 outrel							   *
*									   *
* locals      :	 s.u.							   *
*									   *
* returns     :	 Errorcode 1 bei Fehlern				   *
*		 Errorcode 0 sonst					   *
*									   *
\**************************************************************************/

void main(argc, argv)
int argc;
char *argv[];
  {
    int	  ppos;					  /* Index '.' in String  */
    char *src;					  /* (fuer Kopieren des	  */
    char *dst;					  /* Dateinamens)	  */



    /* Kommandozeilenargumente testen und verarbeiten */

    if (argc !=	2 && argc != 3)	usagerr();	  /* Anz. Argumente ok ?  */
    ++argv;					  /* (1. Angabe)	  */
    src	= *argv;				  /* Quelldateiname	  */
    dst	= srcfn;				  /* in	Uppercase	  */
    while ((*dst++ = *src++) !=	0x00);		  /* kopieren		  */
    if (argc ==	3)				  /* Option vorhanden ?	  */
      {
	++argv;					  /* (2. Angabe)	  */
	if (	strcmp(*argv,"-l") == 0		  /* ja, muss "-l" sein,  */
	     ||	strcmp(*argv,"-L") == 0
	   )
	  list = TRUE;				  /* dann Listing,	  */
	else
	  usagerr();				  /* sonst Fehler	  */
      }



    /* Startmeldung ausgeben */

    if (!list)
     {
      puts("\nNORD><LINK Z80 Crossassembler Version 1.1");
      puts(__DATE__);
      puts("Copyright (C) 1990 by Michael Roehner, DC4OX");
      puts("ported to MS-DOS by df2au");
      puts("ported to Linux by dl4sdc");
      puts("distribution permitted for non-commercial use in amateur radio\n");
     }

    /* Quelldateinamen aufbauen	*/

    ppos = strlen(srcfn) - 1;
    while (    ppos > 0
	    && srcfn[ppos] != '\\'
	    && srcfn[ppos] != '/'
	    && srcfn[ppos] != '.' ) --ppos;
    if (srcfn[ppos] != '.') strcat (srcfn,".MAC");



    /* Objektdateinamen	aufbauen */

    strcpy(objfn,srcfn);
    ppos = strlen(objfn) - 1;
    while (    ppos > 0
	    && objfn[ppos] != '\\'
	    && objfn[ppos] != '/'
	    && objfn[ppos] != '.' ) --ppos;
    if (objfn[ppos] == '.') objfn[ppos]	= '\0';
    strcat (objfn,".rel");



    /* Speicherplatz zuweisen (keine dynamische	Zuweisung) */

    if (!(st = malloc(STBUFSZ)))
      fatal("Not enough	memory for symbol table","");
    stend = st + (unsigned) STBUFSZ;
    for	(stptr = st; stptr != stend; *stptr++ =	0) ;
    if (!(stp =	(char **)malloc((STMAX * PTRSZ))))
      fatal("Not enough	memory for symbol table	pointer","");
    if (!(mt = mtnext =	malloc(MTBUFSZ)))
      fatal("Not enough	memory for macro table","");
    mtend = mt + (unsigned)(STBUFSZ - MAXLINE);



    /* Dateien oeffnen / erzeugen */

    if (!(srcfile = fopen(srcfn,"rt")))		  /* Quelldatei	oeffnen	  */
      fatal("Unable to open: ",srcfn);		  /* Fehler -> Abbruch	  */
    if (!(outrel = fopen(objfn,"wb")))		  /* Objektdatei oeffnen  */
      fatal("Unable to create: ",objfn);	  /* Fehler -> Abbruch	  */



    /* Assemblerdurchlaeufe */

    pass1();				     /*	Symboltabelle aufbauen	  */
    rewind(srcfile);			     /*	noch einmal von	vorne	  */
    pass2();				     /*	Objektcode generieren	  */
    fclose(srcfile);			     /*	Dateien	schliessen	  */
    fclose(outrel);
    if (err)				       /* Fehler ?		  */
      {
	if (!list) puts("Errors	detected\n");  /* ja, Meldung und	  */
	exit(1);			       /* Fehlerabbruch		  */
      }
    printf("\nDone.\n\n");
    exit(0);
  }





/**************************************************************************\
*									   *
* action      :	 Erster	Assemblerdurchgang, Pass 1, Aufbau der		   *
*		 Symboltabelle und des Makrobuffers.			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID pass1() {
  printf("\nPass 1...");
  dopass();				/* Pass	1 durchfuehren		  */
  }





/**************************************************************************\
*									   *
* action      :	 Zweiter Assemblerdurchgang, Pass 2, Erzeugen des	   *
*		 Objektcodes und ggf. des Listings.			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 outrel	 -  Ausgabedatei				   *
*									   *
* r/w globals :	 pass	 -  wird auf 2 gesetzt				   *
*		 err	 -  wird bei Schreibfehlern auf	Ausgabedatei auf   *
*			    YES	gesetzt					   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID pass2() {
  printf("\nPass 2...");
  putname();				/* Modulname deklarieren	  */
  putent();				/* Einsprungpunkte deklarieren	  */
  putsz();				/* Programmgroesse deklarieren	  */
  pass = 2;				/* Pass	2 einleiten		  */
  dopass();				/* und durchfuehren		  */
  putexs();				/* externe Symbole deklarieren	  */
  putend();				/* Programmende	deklarieren	  */
  if (ferror(outrel)) err = YES;	/* REL-Datei-Schreibfehler ?	  */
  }





/**************************************************************************\
*									   *
* action      :	 Pass 1	und Pass 2 durchfuehren.			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 \							   *
*		   alle	globalen Variable, s.u.				   *							     *
* r/w globals :	 /							   *
*									   *
* locals      :	 mop	-  aktueller Makroop. (MACRO/REPT/ENDM/CALL/NO)	   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID dopass() {

  int		mop;

  cloc	 =
  dloc	 =
  mlnext =
  lpage	 =
  lin	 =
  loc	 = 0;				   /* alles zuruecksetzen	*/
  seg	 = PREL;			   /* default Startsegment	*/

  lline	= 100;				   /* Seitenvorschub erzwingen	*/
  eom =	NO;				   /* noch nicht am Modulende	*/
					     /*	1. Zeile holen		  */
  if (!fgets(line,MAXLINE,srcfile))	   /* sollte schon da sein...	*/
    fatal("Empty Sourcefile","");	     /*	Programmabbruch		  */
  LOOP {				    /* fuer jede Zeile		 */
    ++lin;				 /* Zeilenzaehler	      */
    lerr = 0;				 /* Zeilenfehlerflag loeschen */
    part1 = NO;				 /* erster Teil	der Zeile     */
					     /*	noch nicht gelistet	  */
    begline();				 /* Listingzeile einleiten    */
    lp = line;				 /* Zeilenanfang	      */
    lp = getsy(lp,NO);			 /* 1. Feld ueberspringen     */
    if (!(mop =	macop()) && gotnam) {	   /* Makrotoken im 2. Feld ?	*/
      lp = skip(1,line);	       /* nein - 1. Feld nehmen	    */
      mop = macop();
      }
    if (defmode) {			 /* Definitionsmodus :	      */
      if (mop == ENDM) defmode = NO;   /* Makroende		    */
      if (pass == 1 || reptmode)       /* Pass 1 oder REPT :	    */
	putmac();		       /* Zeile	in Makrotabelle	    */
      if (reptmode && !defmode)	       /* Ende REPT-Definition,	    */
	if (!reptno) {			/* wenn	kein REPT	     */
					     /*	auszufuehren,		  */
	  reptmode = NO;	     /*	nichts wiederholen und	  */
	  mtnext = mtrept;	     /*	REPT-Buffer zurueckgeben  */
	  }
	else {
	  expmode = YES;	     /*	sonst REPT expandieren	  */
	  mtptr	= mtrept;
	  }
      }
    else
      switch (mop) {			 /* kein Definitionsmodus :   */
	case CALL :		       /* Makroaufruf :		    */
	  expmode = YES;	       /* Expansionsmodus ein	    */
	  putparm();		       /* Parameter retten	    */
	  dolabel();		       /* Label	verarbeiten	    */
	  break;

	case MACRO :		       /* Makrodefinitionsstart	:   */
	  defmode = YES;	       /* Definitionsmodus ein	    */
	  if (pass == 1) newmac();     /* neuen	Makro anlegen	    */
	  break;

	case REPT :		       /* Repeatdefinitionsstart :  */
	  reptmode = defmode = YES;    /* Definitionsmodus ein	    */
	  mtrept = mtnext;	       /* Stelle im Buffer merken   */
	  lp = skip(2,lp);	       /* REPT ueberspringen	    */
	  onexpr();		       /* Ausdruck holen/auswerten  */
	  reptno = field;	       /* und in Repeat-Zaehler	    */
	  dolabel();		       /* moegliches Label holen    */
	  break;

	case ENDM :		       /* Makroexpansion am Ende,   */
	  if (reptmode)	{	       /* bei REPT-Expansion	    */
	    if (--reptno) {			 /* solange zu wiederholen .. */
	      mtptr = mtrept;	   /* ... von vorne		*/
	      *lp++ = '\n';	   /* ENDM nicht in Listing	*/
	      *lp = '\0';
	      }
	    else {			/* alle	Wiederh. geschafft,  */
	      reptmode =	   /* REPT vergessen,		*/
	      expmode  = NO;	   /* nichts mehr expandieren	*/
	      mtnext = mtrept;	   /* REPT-Buffer zurueckgeben	*/
	      }
	    }
	  else			       /* keine	REPT-Expansion :    */
	    expmode = NO;	       /* Expansionsmodus verlassen */
	  break;

	  default :			 /* sonst,		      */
	    iloc = loc;			 /* Positionszaehler fuer $   */
	    if (expmode	&& !reptmode)	 /* im MACRO-Expansionsmodus  */
	      replace();		 /* Makroparameter ersetzen   */
	    dolabel();			 /* Label verarbeiten	      */
	    if (!domach()) doasm();	 /* Maschineninstruktion oder */
	    break;			   /* Pseudoop verarbeiten	*/
	  } /* end switch (mop)	*/

	endline();			     /*	Listingzeile abschliessen */

	if (pass == 2)			     /*	Pass 2 :  Fehler ent-	  */
	  gripe();			     /*	sprechend lerr ausgeben	  */

	if (expmode)			     /*	Expansionsmodus	naechste  */
	  getmac();			     /*	Zeile aus Makrobuf. holen */

      else {
	if (eom) break;			 /* nur	wenn nicht END war    */
	if (!fgets(line,MAXLINE,srcfile))/* falls am Ende ohne END,   */
	  fatal("Missing END","");	 /* Programmabbruch	      */
	}
    } /* end LOOP */

  if (defmode) {		       /* Source zuende, innerhalb  */
    err	= YES;				 /* kein ENDM :	 Fehler,      */
    puts("- Missing ENDM");		 /*		 melden	      */
    }

  }





/**************************************************************************\
*									   *
* action      :	 Programmabbruch mit Bedienungshinweis.			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 (Programmabbruch mit Errorcode	1)			   *
*									   *
\**************************************************************************/

VOID usagerr()
  {
    fatal("Usage is 'mac <filespec> [-l]'","");
  }



/* Ende	von MAC1.C */
