/********************************************************/
/*							*/
/*	    Q/C	Compiler Standard Definitions		*/
/*							*/
/*     Copyright (c) 1985 Quality Computer Systems	*/
/*		 All Rights Reserved			*/
/*							*/
/*			07/13/85			*/
/********************************************************/

#ifdef NOARGS
#define __ARGS(x) ()
#else
#define __ARGS(x) x
#endif

#define	TRUE		1
#define	FALSE		0

/*
 *	Define code generation parameters
 */

#define	FILE_ULINE	'1'	/* What	'_' changes to when building a file
				   name	using the library (-L) option */
#define	ASMULINE	'@'	/* What	'_' changes to in assembler output */
#define	ASMRTS		'?'	/* Used	as prefix or suffix on all assembler */
				/* names to prevent colliding with keywords */

#define	MAXREG		5	/* max number of register variables */
#define	MAXINCL		3	/* max depth for nested	#include */
#define	MAXIF		6	/* max depth for nested	#if */
#define	INITSIZE	128	/* largest array initialized without -I	*/

#define	CODE		1	/* values for segtype:	code segment */
#define	DATA		2	/*			data segment */
#define	GENCODE		1	/* generating code for a plain expression*/
#define	GENARG		2	/* generating args in a	func call */
#define	PREG		1	/* the primary register	(HL) */
#define	SREG		2	/* the secondary register (DE) */
#define	WREG		2	/* the work register (DE) */
#define	SFP		3	/* the stack frame pointer (BC)	*/
#define	XREG		4	/* Z80 index register (IX) */

#define	EQ		1	/* codes for binary operator routine */
#define	NE		2
#define	LE		3
#define	GE		4
#define	LT		5
#define	GT		6

/*	Define peephole	optimizer constants	*/

#define	JUMP		1
#define	JUMPTRUE	2
#define	JUMPFALSE	3
#define	PREGSYM		4
#define	ADDRSFP		5

#define	TESTED		1
#define	REGH		2
#define	REGL		3

/*
 * Define "smart" switch for compiler options which can
 * be turned on	or off by special comments of the form:
 *
 *	/ * $ +|-S(witch)
 *
 * from	within the C program.
 */
struct _switch {
	int	is_on;		/* current switch setting */
	int	enabled;	/* was option requested	in compile */
	};

/*	Define the type	table		*/

struct typeinfo	{
	char	 t_code;	/* int,	ptr to,	array of, etc. */
	int	 t_size;	/* size	of one thing of	this type */
	unsigned t_refs;	/* # of	vars pointing to this entry */
	union baseinfo {
	    struct memtab	/* pointer to the member table when */
		*memlist;	/*	t_code == T_STRUCT or ... */
	    struct typeinfo	/* pointer to the base type such as */
		*p_type;	/*	'int' of 'pointer of int' */
		} t_base;
	struct typeinfo		/* next	type entry for this t_code */
		*t_next;
	};

struct parsestack {
	struct typestack {
		char	 t_code;
		int	 t_size;
		} *curr_ptr, stack[30];
	};

/*
 *	Type codes:
 *
 *	T_xxx is the code for the type 'xxx'.
 *	S_xxx is the size, in bytes, of	one item of type 'xxx'.
 *
 *	T_NONE is assumed to be	non-negative, and less than all	other
 *		type codes.
 *	All types between T_NONE and T_SIMPLE are 'simple' types:
 *		they may be returned by	functions.  All	other types
 *		are 'non-simple' and may not be	returned by functions.
 */

#define	T_NONE		0

#define	T_CHAR		1
#define	S_CHAR		1
#define	T_SHORT		2
#define	S_SHORT		2
#define	T_INT		3
#define	S_INT		2
#define	T_UNSIGNED	4
#define	S_UNSIGNED	2
#define	T_LONG		5
#define	S_LONG		4
#define	T_FLOAT		6
#define	S_FLOAT		4
#define	T_DOUBLE	7
#define	S_DOUBLE	4
#define	T_PTR		8
#define	S_PTR		2

#define	T_SIMPLE	T_PTR

#define	T_LABEL		9		/* statement label */
#define	T_STRUCT	10
#define	T_UNION		11
	/* The following are all compound types	*/
#define	T_ARRAY		12		/* array of ...	*/
#define	T_FUNC		13		/* function returning ... */
#define	T_MAX		14

#define	TO_CUR		1
#define	FRM_CUR		0

#define	KEEP		1
#define	NO_KEEP		0

/*	Define the symbol table	parameters	*/

#define	STARTLOC (symtab+(nsym-1))
#define	NAMEMAX	 14 /* ACZ : was 8 before */
#define	NAMESIZE (NAMEMAX+1)

/*
 *	Define symbol table
 */

struct st {
	char	 st_sc;		/* storage class */
	struct typeinfo		/* variable type */
		 *st_type;
	int	 st_info;	/* stack offset	for local auto	*/
				/* register # for local	register*/
				/* offset for structure	member	*/
				/* label # for local static	*/
				/* label # for statement label	*/
	char	 st_name[NAMESIZE];
	char	 st_idset;	/* tag/member or variable */
	};

/*
 *		Define hash index into symbol table
 */

struct st_hash {
	struct st	*sym;		/* corresponding symbol	table entry */
	struct st_hash	*next_sym;	/* next	entry on synonym chain */
	};

/*	Define possible	entries	for "st_idset"	*/
#define	ID_VAR		1
#define	ID_STRUCT	2

/*	Define possible	entries	for "st_sc"  */

#define	SC_NONE		0
#define	SC_GLOBAL	1	/* unspecified external	*/
#define	SC_ST_GLB	2	/* static ... (global) */
#define	SC_EXTERN	3	/* extern ... */
#define	SC_STATIC	4	/* static ... (local) */
#define	SC_AUTO		5	/* auto	... or unspecified local */
#define	SC_ARG		6	/* like	auto, but for arg on stack */
#define	SC_REG		7	/* register ...	*/
#define	SC_MEMBER	8	/* member of, or tag for, structure */
				/* distinguished by st_ident */
#define	SC_TYPE		9	/* typedef ...,	or struct xxx */

/*	Define possible	entries	for "st_info"	*/

#define	DECL_GLB	1
#define	DECL_LOC	2

/*	Define structure member	table		*/

struct memtab {
	struct st     *sym;	/* pointer to symbol table entry */
	struct memtab *nextmem;	/* next	member or NULL */
	};

/*	Define expression operand for parser	*/

struct operand {
	char		op_load;
	struct st	*op_sym;
	long		op_val;
	struct typeinfo	*op_type;
	};

/*	Define possible	entries	for "op_load"	*/

#define	EXPRESSION	0	/* an operand whose value is in	preg */
#define	LOADVALUE	01	/* an op whose value needs loading */
#define	LOADADDR	02	/* an op whose address needs loading */
#define	LVALUE		04	/* an lvalue */
#define	CONSTANT	010	/* a constant expression */
#define	CONSTADDR	020	/* a constant address expression */
#define	CONSTOFF	040	/* a stack var with constant offset */
/* This	is the result of constant assignment statements	*/
#define	ASGCONST	0100

/*	Define the switch/loop queue		*/

struct swq {
	int loop;
	int exit;
	};

/*	Define the switch case table		*/

struct case_table {
	int value;
	int label;
	};

/*
 *	Define hash index into the macro (#define) pool
 */

struct mp_hash {
	char		name[NAMESIZE];	/* macro name */
	char		*def;		/* ptr to replacement text */
	struct mp_hash	*next_mac;	/* next	entry on synonym chain */
	};

/*
 *	Define the input line
 */

#define	LINESIZE 128
#define	LINEMAX	(LINESIZE-1)

/*	Define conditional compilation parameters	*/

#define	IGNORE	0
#define	PROCESS	1
#define	SKIP	2



extern char *clrsbrk();
extern char *farmalloc();
extern long farcoreleft();
extern FILE *fopena();
extern FILE *fopenb();

/* function prototypes */

void	zeromem __ARGS((void));
void	inittable __ARGS((void));
void	parse __ARGS((void));
void	dumplits __ARGS((void));
void	summary __ARGS((void));
int	isdecl __ARGS((int *, struct typeinfo **));
void	declglb __ARGS((int, struct typeinfo *));
void	declloc __ARGS((int, struct typeinfo *));
int	arraysize __ARGS((void));
struct typeinfo		*decltag __ARGS((int));
struct typeinfo		*declvar __ARGS((char *, struct	typeinfo *, int, int *));
int	needreg __ARGS((struct typeinfo	*));
void	initvar __ARGS((struct typeinfo	*));
void	initaggreg __ARGS((struct typeinfo *, int));
int	initstr __ARGS((struct typeinfo	*));
void	initchar __ARGS((void));
void	initbyte __ARGS((int));
void	initword __ARGS((int));

int	isfunc __ARGS((char *, struct typeinfo *, int));
void	procarg __ARGS((void));
void	declarg __ARGS((int, struct typeinfo *));
void	prologue __ARGS((char *));
void	epilogue __ARGS((char *));

void	compound __ARGS((void));
void	statement __ARGS((void));
int	chklabel __ARGS((void));
void	ns __ARGS((void));
void	doif __ARGS((void));
void	dowhile __ARGS((void));
void	dofor __ARGS((void));
void	dodowhile __ARGS((void));
void	doreturn __ARGS((void));
void	dobreak __ARGS((void));
void	docont __ARGS((void));
void	doswitch __ARGS((void));
int	docase __ARGS((struct case_table *, int	*));
void	dogoto __ARGS((void));
void	testexp __ARGS((void));

void	expression __ARGS((int));
void	heir1 __ARGS((struct operand *,	int));
void	heir2 __ARGS((struct operand *,	int));
void	heir3 __ARGS((struct operand *));
void	heir4 __ARGS((struct operand *));
void	heir5 __ARGS((struct operand *));
void	heir6 __ARGS((struct operand *));
void	heir7 __ARGS((struct operand *));
void	heir8 __ARGS((struct operand *));
void	heir9 __ARGS((struct operand *));
void	heir10 __ARGS((struct operand *));
void	heir11 __ARGS((struct operand *));
void	heir12 __ARGS((struct operand *));
void	heir13 __ARGS((struct operand *));
void	heir14 __ARGS((struct operand *));
void	heir15 __ARGS((struct operand *));
int	ismember __ARGS((struct	operand	*));
void	primary __ARGS((struct operand *));
void	unarray __ARGS((struct operand *));
void	setconstaddr __ARGS((struct operand *));
void	initop __ARGS((struct operand *, struct	st *, struct typeinfo *, int));
int	needlval __ARGS((struct	operand	*));

int	chkeyword __ARGS((void));
int	chkdecl __ARGS((void));
int	chktype __ARGS((void));
int	chksc __ARGS((void));
struct typeinfo	*gettype __ARGS((void));
struct typeinfo	*istypedef __ARGS((void));
int	getsc __ARGS((void));
struct typeinfo	*abstdecl __ARGS((void));
struct typeinfo	*loadtype __ARGS((struct typeinfo *, struct parsestack *));
void	dodecl __ARGS((struct parsestack *, char *));
void	pushtype __ARGS((struct	parsestack *, int, int));
int	symname __ARGS((char *));
char	*chksym __ARGS((char *));
int	addlabel __ARGS((char *, int));
int	getlabel __ARGS((void));
void	addswq __ARGS((struct swq *));
void	delswq __ARGS((void));
struct swq	*getswq __ARGS((void));
void	addcase __ARGS((struct case_table *, struct case_table *));
struct case_table	*delcase __ARGS((void));
struct mp_hash	*findmac __ARGS((char *));
void	addmac __ARGS((char *));
void	delmac __ARGS((void));
struct mp_hash	*isdef __ARGS((void));
int	constexp __ARGS((struct	operand	*));
int	constant __ARGS((struct	operand	*));
int	hex __ARGS((long *));
int	octal __ARGS((long *, char *));
int	decimal __ARGS((long *,	char *));
int	charconst __ARGS((long *));
int	string __ARGS((struct operand *));
int	getstring __ARGS((void));
char	escape __ARGS((void));
struct st	*findglb __ARGS((char *));
struct st	*findloc __ARGS((char *));
struct st	*findtag __ARGS((char *));
struct st	*findvar __ARGS((struct	st *, struct st	*, char	*, int));
struct st	*addglb __ARGS((char *,	struct typeinfo	*, int,	int, int));
struct st	*addloc __ARGS((char *,	struct typeinfo	*, int,	int, int));
struct st	*addvar __ARGS((struct st *, char*, struct typeinfo *, int, int, int));
struct st	*find_sym __ARGS((char *, int));
void	add_sym __ARGS((char *,	struct st *));
int	hash __ARGS((char *, int));
struct memtab	*addmem __ARGS((struct st *, struct typeinfo *,	struct memtab *));
void	delmem __ARGS((struct memtab *));
void	inittypes __ARGS((void));
struct typeinfo	*addtype __ARGS((int, union baseinfo *,	int));
struct typeinfo	*uniquetype __ARGS((int, union baseinfo	*, int));
void	reftype __ARGS((struct typeinfo	*));
void	unreftype __ARGS((struct typeinfo *));
struct typeinfo	*ptrto __ARGS((struct typeinfo *));
struct typeinfo	*findtype __ARGS((int, union baseinfo *, int));
struct typeinfo	*alloctype __ARGS((int,	union baseinfo *, int));
void	scraptype __ARGS((struct typeinfo *));
void	freetype __ARGS((struct	typeinfo *, struct typeinfo *));
int	typecollect __ARGS((void));
void	deref __ARGS((struct operand *));
void	settype __ARGS((struct operand *, struct operand *));
void	chtype __ARGS((struct operand *, struct	typeinfo *));
int	basesize __ARGS((struct	operand	*));
struct typeinfo	*isptr __ARGS((struct operand *));
int	islval __ARGS((struct operand *));
int	isconstant __ARGS((struct operand *));
int	isloaded __ARGS((struct	operand	*));
int	isaddr __ARGS((struct operand *));
int	is32bit __ARGS((struct typeinfo	*));
struct typeinfo	*comptyp __ARGS((struct	operand	*, struct operand *));
int	cnvreq __ARGS((struct typeinfo *, struct typeinfo *));

void	gentrace __ARGS((int, char *));
int	binop __ARGS((struct operand *,	int, struct operand *, void (*)()));
int	specialcase __ARGS((struct operand *, int, struct operand *));
void	doadd __ARGS((struct operand *,	struct operand *));
void	addto __ARGS((struct operand *,	int));
void	increment __ARGS((struct operand *, int));
void	dosub __ARGS((struct operand *,	struct operand *));
void	domult __ARGS((struct operand *, struct	operand	*));
void	multconst __ARGS((int));
void	callfunction __ARGS((struct operand *));
void	loadarg __ARGS((int));
void	reverse __ARGS((char *));
void	argtoreg __ARGS((struct	st *, int));
int	loadops __ARGS((struct operand *, struct operand *));
void	rvalp __ARGS((struct operand *));
void	rvalpsav __ARGS((struct	operand	*));
void	rvals __ARGS((struct operand *));
void	ld_float __ARGS((struct	operand	*));
void	load __ARGS((struct operand *));
void	cnvrtlt __ARGS((struct operand *, int));
void	store __ARGS((struct operand *,	int));
void	indirect __ARGS((struct	operand	*));
void	sindirect __ARGS((struct operand *));
void	findirect __ARGS((struct operand *));
void	addconst __ARGS((long));
char	fprfx __ARGS((void));
void	ccomp __ARGS((char *));
void	cord __ARGS((char *, int));
void	clogic __ARGS((char *));
void	cmult __ARGS((void));
void	cneg __ARGS((void));
void	savepreg __ARGS((void));
void	savetos __ARGS((void));
void	lopimmed __ARGS((struct	operand	*, int));
void	lsimmed __ARGS((int, int));
void	outname __ARGS((char *));
void	printlabel __ARGS((int));
void	peeplabel __ARGS((int));
void	loadpeep __ARGS((char *));
void	dumpeep __ARGS((void));
void	clearpeep __ARGS((void));

void	header __ARGS((void));
void	shell __ARGS((int));
void	trailer __ARGS((void));
void	sextout __ARGS((void));
void	lextout __ARGS((void));
void	fextout __ARGS((void));
void	comment __ARGS((void));
void	doasm __ARGS((void));
void	doasgn __ARGS((struct operand *, struct	operand	*, int,	int));
void	asgindirect __ARGS((long, int));
void	autoload __ARGS((struct	operand	*));
void	getaddr __ARGS((struct operand *));
void	qgetmem __ARGS((struct operand *));
void	putstk __ARGS((struct operand *, int));
void	putmem __ARGS((struct operand *, int));
void	put32 __ARGS((int));
void	prtsymname __ARGS((struct st *,	int));
void	prtname __ARGS((struct st *));
void	xreg __ARGS((int));
char	*regname __ARGS((int));
void	swap_reg __ARGS((void));
void	pimmed __ARGS((void));
void	simmed __ARGS((void));
void	ldmemi __ARGS((int));
void	set1 __ARGS((void));
void	cpush __ARGS((int));
void	cpop __ARGS((int));
void	restsfp __ARGS((void));
void	swapstk __ARGS((void));
void	callib __ARGS((char *));
void	ccall __ARGS((char *));
void	calladdr __ARGS((int));
void	jump __ARGS((int));
void	jumpcond __ARGS((int, int));
void	dotest __ARGS((void));
void	linelabel __ARGS((int));
void	glblabel __ARGS((char *, int));
void	dataseg __ARGS((void));
void	codeseg __ARGS((void));
void	declext __ARGS((char *));
int	definit __ARGS((int));
void	defbyte __ARGS((void));
void	defstorage __ARGS((void));
void	defword __ARGS((void));
void	clearstk __ARGS((int));
void	doublereg __ARGS((void));
void	cadd __ARGS((void));
void	cinc __ARGS((void));
void	cdec __ARGS((void));
void	cret __ARGS((void));

void	getinfil __ARGS((void));
void	getoutfil __ARGS((void));
void	closeout __ARGS((void));
void	getfilename __ARGS((char *, char *, char *));
void	doinclude __ARGS((void));
void	getline __ARGS((void));
void	outbyte __ARGS((int));
void	outstr __ARGS((char *));
int	chkabort __ARGS((void));
void	newline __ARGS((void));
void	ot __ARGS((char	*));
void	ol __ARGS((char	*));
void	nl __ARGS((void));
int	match __ARGS((char *));
int	matchc __ARGS((int));
int	nextc __ARGS((int));
int	amatch __ARGS((char *));
void	blanks __ARGS((void));
void	junk __ARGS((void));
int	endst __ARGS((void));
int	ch __ARGS((void));
int	nch __ARGS((void));
int	gch __ARGS((void));
void	putback __ARGS((int));
void	kill_line __ARGS((void));
void	outdec __ARGS((int));
void	chkcommand __ARGS((void));
void	preprocess __ARGS((void));
void	dopreif __ARGS((int));
void	procline __ARGS((void));
void	keepch __ARGS((int));
void	fatal __ARGS((char *));
void	io_err __ARGS((int, char *));
void	error __ARGS((int));
void	errname __ARGS((int, char *));
void	prtfile __ARGS((void));
void	prtmesg __ARGS((int));
void	errptr __ARGS((void));
void	errpause __ARGS((void));
void	fltoperr __ARGS((void));
void	out_err __ARGS((void));
void	matcherr __ARGS((void));
void	scerr __ARGS((void));
void	typerr __ARGS((void));
void	illname __ARGS((void));
void	multidef __ARGS((char *));
void	tagerr __ARGS((char *));
void	needpunc __ARGS((int));
void	keyerr __ARGS((char *));
void	sizerr __ARGS((void));
void	initerr __ARGS((void));
void	cexperr __ARGS((void));
/*
int	issym __ARGS((char));
int	issymf __ARGS((char));
*/
int	chks __ARGS((char *));
int	streq __ARGS((char *, char *));
int	astreq __ARGS((char *, char *, int));
void	skip __ARGS((void));

void	init __ARGS((void));
void	getoptions __ARGS((int, char **));
int	gethex __ARGS((void));
void	usagerr __ARGS((void));
void	getspace __ARGS((void));

void initsbrk __ARGS((void));
int imin __ARGS((int, int));
int imax __ARGS((int, int));
void kbclear __ARGS((void));
int inkey __ARGS((void));

/* end of QC.H */
