/*
**	File: LMISC.C
**
**	QCX Library
**
**	Misc. non-standard QCX routines.
**
**	This I/O library assumes a CP/M 2.x target operating system.
**	Some of these routines are included for backward compatibility
**	with UNIX Version 7.
**
**	Copyright (c) 1984, 1985 Quality Computer Systems
**	All rights reserved.
**
**	Published by:	The Code Works
**			Box 62136
**			Santa Barbara, CA 93160
**
**	Modified: 4Apr85 rj
**		 23Jul85 jc
*/

#include "stdio.h"
#include "ctype.h"

/* Library parameters */


#define	TRUE	1
#define	FALSE	0
#define MAXARG	20	/* max no. of command line arguments */
#define MAXLINE 81	/* maximum size of console input line */
#define MAXSCAN 128	/* maximum size of string in _scanf %[] format */
#define NFILES	10	/* max no. of files that can be open */
#define SECSIZE 128	/* size of CP/M record */

/* Miscellaneous CP/M constants */

#define CPMFCBSIZE 36
#define DBUFF	0x80
#define CPMERR	0xFF
#define CPMEOF	0x1A
#define CTLC	0x3	/* ^C */

/* CP/M function numbers */

#define CONIN	1
#define CONOUT	2
#define LSTOUT	5
#define PRINTS	9
#define CONBUF	10
#define CONSTAT 11
#define OPENF	15
#define CLOSEF	16
#define DELETEF 19
#define MAKEF	22
#define SETDMA	26
#define READR	33
#define GTFSIZE 35
#define WRITERZ 40

/* I/O flag masks */

#define READ	01
#define WRITE	02
#define APPEND	04
#define BINARY	010
#define BUF	020
#define USERBUF 040
#define FEOF	0100
#define FERR	0200

/* File control block for low-level I/O */

struct fcb {
	char	flag;
	unsigned file_size;
	struct cpmfcb {
		char	drive;
		char	filename[8];
		char	filetype[3];
		char	extent;
		int	reserved;
		char	rec_cnt;
		char	dir[16];
		char	curr_rec;
		unsigned rand_rec;
		char	overflow;
		} cpm_fcb;
	};

/*: index -- find the first occurrence of a given character */
/*		Note: UNIX V7 name calling the new C standard name */
char *index(s, c)
char *s;
	{
	char *strchr();

	return strchr(s, c);
	}

/*: rindex -- find the last occurrence of a given character */
/*		Note: UNIX V7 name calling the new C standard name */
char *rindex(s, c)
char *s;
	{
	char *strrchr();

	return strrchr(s, c);
	}

/*: strmov -- copy second string into first string & point to end */
char *strmov(s1, s2)
char *s1, *s2;
	{
	register char *p;

	p = s1;
	while (*p = *s2++)	/* copy 2nd string into 1st */
		++p;
	return p;
	}

/*: unlink -- Delete (erase) CP/M file */
/*		Note: UNIX V7 name calling the new C standard name */
unlink(filename)
char *filename;
	{
	return remove(filename);
	}

/*: _exit -- Immediate exit to CP/M (UNIX V7 equivalent of abort) */
_exit()
	{
	bdos1(0);
	}

/*: peek -- get the byte at address */
peek(address)
char *address;
	{
	return (*address & 0xFF);
	}

/*: poke -- put value in the byte at address */
poke(address, value)
char *address, value;
	{
	int oldval;

	oldval = (*address & 0xFF);
	*address = value;
	return oldval;
	}

/*: wpeek -- get the word at address */
wpeek(address)
int *address;
	{
	return *address;
	}

/*: wpoke -- put value in the word at address */
wpoke(address, value)
int *address;
	{
	int oldval;

	oldval = *address;
	*address = value;
	return oldval;
	}

/*: seekr -- Position file for random I/O */
seekr(fd, offset, mode)
unsigned offset;
	{
	struct fcb *_gfcb(), *fcb;
	register struct cpmfcb *cpmfcb;

	if ((fcb = _gfcb(fd)) == NULL)
		return -1;
	cpmfcb = &fcb->cpm_fcb;
	switch (mode) {
	case 2: 			/* from EOF */
		cpmfcb->rand_rec = fcb->file_size;
	case 1: 			/* from current rec */
		cpmfcb->rand_rec += offset;
		break;
	case 0: 			/* from beginning of file */
		cpmfcb->rand_rec = offset;
		break;
	default:
		return -1;
		}
	fcb->flag &= ~FEOF;		/* if we saw EOF, forget it */
	return 0;
	}

/*: tellr -- Report current record number in file */
unsigned tellr(fd)
	{
	struct fcb *fcb, *_gfcb();

	if ((fcb = _gfcb(fd)) == NULL)
		return 0;
	else
		return (fcb->cpm_fcb.rand_rec);
	}


/*: getw -- Read a word from a file */
getw(fp)
FILE *fp;
	{
	register unsigned low, high;

	if ((low=getc(fp)) == EOF || (high=getc(fp)) == EOF)
		return EOF;
	else
		return ((high << 8) | low);
	}

/*: putw -- Write a word to a file */
putw(word, fp)
unsigned word;
FILE *fp;
	{
	if (putc(word & 0xFF, fp) == EOF || putc(word >> 8, fp) == EOF)
		return EOF;
	else
		return (word);
	}



/*: setbsize -- Set size of user-provided buffer */
setbsize(fp, size)
FILE *fp;
	{
	if (fp->_flag & BUF)	/* system buffer allocated? */
		return;
	else if (size /= SECSIZE)
		fp->_bufsize = size * SECSIZE;
	else			/* can't use less than one sector */
		fp->_flag &= ~USERBUF;
	}

/*: _rshell -- Parse command line and open standard files (does redirection) */
_rshell()
	{
	static	argc,
		argv[MAXARG],
		argflag = 0,	/* not in an argument */
		cmdcnt,
		c;
	static char
		*cmdline = 0x80, /* CP/M command line at 80H */
		*pargs,
		*ptr;
	char args[128], infile[15], outfile[15];
	FILE *fopen();

	cmdcnt = *cmdline++;	/* first byte is length */
	pargs = args;		/* point to args */
	*pargs =		/* set command = null string */
	infile[0] =		/* no redirection of standard files */
	outfile[0] = '\0';
	argv[argc++] = pargs;	/* record pointer to 1st arg */
	while (cmdcnt--) {
		c = *cmdline++;
		switch (argflag) {
		case 0: 		/* not in an arg */
			if (isspace(c))
				continue; /* still not in an arg */
			if (argc == MAXARG) {
				fputs("Too many command line args\n", stderr);
				exit(1);
				}
			argflag = c;	/* set type of arg */
			if (c == '<')	/* redirect stdin */
				ptr = infile;
			else if (c == '>') /* redirect stdout */
				ptr = outfile;
			else {		/* just an arg */
				argv[argc++] = pargs + 1;
				if (c != '"')	/* if not a string */
					break;	/* record char */
				}
			continue;	/* skip this character */
		case '<':		/* record redirection */
		case '>':		/* file name */
			if (isspace(c))
				c = argflag = 0; /* end of name */
			*ptr++ = c;
			continue;
		case '"':		/* in a string */
			if (c == '"')
				c = argflag = 0; /* end of string */
			break;
		default:
			if (isspace(c))
				c = argflag = 0; /* end of arg */
			break;
			}
		*++pargs = c;
		}
	if (argflag == '<' || argflag == '>') /* end last arg */
		*ptr = '\0';		/* a redirected filename */
	else
		*++pargs = '\0';	/* a normal arg */
	if (infile[0] != '\0') {	/* check for redirection */
		if ((stdin = fopen(infile, "r")) == NULL)
			cantopen(infile);
		}
	if (outfile[0] != '\0') {
		if ((stdout = fopen(outfile, "w")) == NULL)
			cantopen(outfile);
		}
	main(argc, argv);
	exit(0);
/* end _rshell */ }

/* end file LMISC.C */
