/*
**	File: LSTDLIB.C
**
**	QCX Library
**
**	D.10 -- General Utilities --
**		D.10.1 string conversion
**		D.10.2 (not yet implemented) random numbers
**		D.10.3 memory management
**		D.10.4 communication with environment
**
**	Copyright (c) 1984, 1985 Quality Computer Systems
**	All rights reserved.
**
**	Published by:	The Code Works
**			Box 62136
**			Santa Barbara, CA 93160
**
**	Modified: 4Apr85 rj
**		 11Sep85 jc
*/

#include "stdio.h"
#include "ctype.h"

#define	TRUE		1
#define	FALSE		0
#define NFILES		10	/* max no. of files that can be open */

/*
** Declare control blocks for buffered files so exit() can close them
*/

extern FILE _iob[];

/*
** Structure for the memory allocator
*/
struct _header {
	struct _header	*_next; /* next free block */
	unsigned	_bsize; /* size of block */
	};
typedef struct _header HEADER;

extern char	*_free; 	/* free space address */
				/* Note: the Microsoft L80 linker puts */
				/* the address of the next available memory */
				/* in the variable $MEMRY. (RMAC uses ?MEMRY) */
				/* the compiler external variable _free */
				/* is at the same location, and so the linker */
				/* does the needed magic */

unsigned	_moat = 1000;	/* moat between heap and stack */
HEADER		_base,		/* empty free space (heap) list */
		*_allocp;	/* last allocated block */

/*: atof -- convert a float string constant to its float value */
float atof(s)
char *s;
	{
	register char *p;
	register c, exp, sign;
	float tenpwr, fvalue;

/* NOTE: do NOT use any floating point constants in this function
 *	 because QCX calls this routine to convert floating point
 *	 constants at run time when it is used as a cross-compiler.
 */
	p = s;
	fvalue = 0;	
	while (isspace(c = *p++))	/* skip leading white space */
		;
	if (c == '-') {
		sign = -1;
		c = *p++;
		}
	else if (c == '+') {
		sign = 1;
		c = *p++;
		}
	else
		sign = 1;
	while (isdigit(c)) {
		fvalue = 10 * fvalue + (c - '0');
		c = *p++;
		}
	if (c == '.') {
		tenpwr = 10;
		c = *p++;
		while (isdigit(c)) {
			fvalue += (c - '0') / tenpwr;
			tenpwr *= 10;
			c = *p++;
			}
		}
	fvalue *= sign;			/* apply correct sign to number */
	if (toupper(c) == 'E') {
		exp = 0;
		c = *p++;
		if (c == '-') {
			sign = -1;
			c = *p++;
			}
		else if (c == '+') {
			sign = 1;
			c = *p++;
			}
		else
			sign = 1;
		while (isdigit(c)) {
			exp = 10 * exp + (c - '0');
			c = *p++;
			if (exp > 100) {	/* artificial limit for */
				exp = 100;	/* while loops below */
				break;
				}
			}
		if (sign > 0)
			while (exp--)
				fvalue *= 10;
		else
			while (exp--)
				fvalue /= 10;
		}
	return fvalue;
/* end atof */	}

/*: atoi -- convert an ASCII decimal string to integer */
atoi(s)
register char *s;
	{
	register sign;
	char *p;

	while (isspace(*s))
		++s;
	sign = 1;
	switch (*s) {
	case '-':
		sign = -1;
	case '+':
		++s;
		break;
		}
	p = s;
	return (sign * _atoi(&p));
	}

/*: _atoi -- internal routine for atoi and _fmt to perform ascii-
 *	     to-decimal conversion and update an associated pointer
 */
_atoi(sptr)
char **sptr;
	{
	register n;
	register char *p;

	n = 0;
	p = *sptr;
	while (isdigit(*p))
		n = 10 * n + *p++ - '0';
	*sptr = p;
	return n;
	}

/*: malloc -- allocate memory */
char *malloc(nbytes)
unsigned nbytes;
	{
	register HEADER *p, *q;
	register unsigned nunits;
	char *sbrk();

	nunits = 1 + (nbytes + (sizeof(HEADER) - 1)) / sizeof(HEADER);
	if ((q = _allocp) == NULL)	/* no free list */
		_base._next = _allocp = q = &_base;
	for (p = q->_next; ; q = p, p = p->_next) {
		if (p->_bsize >= nunits) {
			if (p->_bsize == nunits)
				q->_next = p->_next;
			else {
				p->_bsize -= nunits;
				p += p->_bsize;
				p->_bsize = nunits;
				}
			_allocp = q;
			return ((char *)(p+1));
			}
		if (p == _allocp) {	/* wrapped around */
			p = (HEADER *) sbrk(nunits * sizeof(HEADER));
			if ((int) p == -1)
				return (char *) NULL;
			p->_bsize = nunits;
			free((char *)(p+1)); /* add to free list */
			p = _allocp;
			}
		}
	}

/*: calloc -- allocate n elements of a given size of memory */
char *calloc(nelements, size)
unsigned nelements, size;
	{
	register char *p, *q;
	char *malloc();

	size *= nelements;
	if ((q = p = malloc(size)) != NULL) {
		while (size--)
			*q++ = 0;
		}
	return p;
	}

/*: free -- free memory allocated with malloc */
free(pblk)
char *pblk;
	{
	register HEADER *p, *q;

	p = (HEADER *)pblk - 1; 	/* point to block header */
	for (q=_allocp; !(p>q && p<q->_next); q=q->_next)
		if (q >= q->_next && (p>q || p<q->_next))
			break;
	if (p+p->_bsize == q->_next) {	/* join to upper neighbor */
		p->_bsize += q->_next->_bsize;
		p->_next = q->_next->_next;
		}
	else
		p->_next = q->_next;
	if (q+q->_bsize == p) { 	/* join to lower neighbor */
		q->_bsize += p->_bsize;
		q->_next = p->_next;
		}
	else
		q->_next = p;
	_allocp = q;
	}

/*: sbrk -- set the memory breakpoint */
char *sbrk(size)
register unsigned size;
	{
	unsigned maxsbrk();

	if (maxsbrk() < size)
		return -1;	/* can't go into moat */
	_free += size;		/* move free space ptr */
	return (_free - size);
	}

/*: moat -- Change size of moat between free space and stack */
unsigned moat(size)
unsigned size;
	{
	unsigned oldsize;

	oldsize = _moat;
	_moat = size;
	return oldsize;
	}

/*: maxsbrk -- Compute memory left to be allocated */
unsigned maxsbrk()
	{
	auto char sp[1];	/* sp is on the stack , so sp = SP */
	return ((sp>_free+_moat) ? sp-(_free+_moat): 0);
	}

/*: exit -- Cleanup routine - close files, etc. */
exit(error)
	{
	register FILE *fp;
	register n;

	for (fp = _iob, n = 0; n < NFILES; ++fp, ++n)
		fclose(fp);		/* close any output files */
	if (error)
		remove("A:$$$.SUB");	/* era submit file, if any */
	bdos1(0);			/* reboot CP/M */
	}

/*: abort -- Immediate exit to CP/M with no cleanup */
abort()
	{
	bdos1(0);
	}

/* end LSTDLIB.C */
