/*
**	File: LSTRING.C
**
**	QCX Library
**	D.11 -- String handling
**
**	Copyright (c) 1984, 1985 Quality Computer Systems
**	All rights reserved.
**
**	Published by:	The Code Works
**			Box 62136
**			Santa Barbara, CA 93160
**
**	Modified: 4apr85 rj
**		 10sep85 jc
*/

#include "stdio.h"

#define	TRUE	1
#define	FALSE	0

/*: strcat -- add second string to end of first string */
char *strcat(s1, s2)
char *s1, *s2;
	{
	register char *p;

	p = s1;
	while (*p)		/* find end of first string */
		++p;
	while (*p++ = *s2++)	/* copy 2nd string to end of 1st */
		;
	return s1;
	}

/*: strncat -- add at most n characters from 2nd string to end of 1st string */
char *strncat(s1, s2, n)
char *s1, *s2;
	{
	register char *p;

	p = s1;
	while (*p)		/* find end of first string */
		++p;
	while (n--) {		/* copy 2nd string to end of 1st */
		if ((*p = *s2++) == '\0')
			break;
		else
			++p;
		}
	*p = '\0';
	return s1;
	}

#ifndef _lint
/*: strcpy -- copy second string into first string */
char *strcpy() {
#asm
;char *strcpy(s1, s2)
;char *s1, *s2;
	LD	HL,2	;get s1
	ADD	HL,SP
	LD	E,(HL)
	INC	HL
	LD	D,(HL)
	INC	HL
	LD	A,(HL)	;get s2
	INC	HL
	LD	H,(HL)
	LD	L,A
	PUSH	DE	;save s1 for return
?scpy1: LD	A,(HL)	;*s2
	LD	(DE),A	;copy to *s1
	OR	A	;check for end-of-string '\0'
	JR	Z,?scpy2
	INC	DE	;move to next char
	INC	HL
	JR	?scpy1
?scpy2: POP	HL	;return s1
	LD	A,H
	OR	L	;set Z flag
#endasm
	}

/*: strlen -- find the length of a string */
strlen() {
#asm
;strlen(s)
;char *s;
	LD	HL,2
	ADD	HL,SP
	LD	E,(HL)	;get s in DE
	INC	HL
	LD	D,(HL)
	LD	HL,0	;init length counter
?slen1: LD	A,(DE)	;get next character
	OR	A	;end of string?
	JR	Z,?slen2
	INC	DE
	INC	HL	;count one more
	JR	?slen1
?slen2: LD	A,H	;length in HL
	OR	L	;set Z flag
#endasm
	}

/*: strcmp -- compare two strings */
strcmp() {
#asm
;strcmp(s1, s2)
;register char *s1, *s2;
	LD	HL,2	;get args
	ADD	HL,SP
	LD	E,(HL)	;s1
	INC	HL
	LD	D,(HL)
	INC	HL
	LD	A,(HL)	;s2
	INC	HL
	LD	H,(HL)
	LD	L,A
?scmp1: LD	A,(DE)	;current char of s1
	SUB	(HL)	;compute *s1-*s2
	JR	NZ,?scmp2	;they don't match here
	OR	(HL)	;end-of-string (0)?
	JR	Z,?scmp2	;they match to end-to-string
	INC	DE	;next char of s1
	INC	HL	;same for s2
	JR	?scmp1	;loop
?scmp2: LD	L,A	;set return value (*s1-*s2)
	RLCA		;sign-extend into HL
	SBC	A,A
	LD	H,A
	OR	L	;set Z flag
#endasm
	}
#endif

/*: strncpy -- copy one string to another for exactly n characters */
char *strncpy(s1, s2, n)
char *s1, *s2;
	{
	char *p;

	p = s1;
	while (--n >= 0) {
		if (*p++ = *s2) /* copy and check for end of s2 */
			++s2;	/* not at end so get next char */
		}
	return s1;
	}

/*: strncmp -- compare one string to another for exactly n characters */
strncmp(s1, s2, n)
register char *s1, *s2;
	{
	if (n <= 0)
		return 0;
	while (--n > 0 && *s1) {
		if (*s1 != *s2)
			break;
		++s1; ++s2;
		}
	return (*s1 - *s2);
	}

/*: strchr -- find the first occurrence of a given character */
char *strchr(s, c)
register char *s;
	{
	while (*s) {
		if (*s == c)
			return s;
		++s;
		}
	return NULL;
	}

/*: strrchr -- find the last occurrence of a character in a string */
char *strrchr(s, c)
register char *s;
	{
	register char *p;

	p = NULL;
	while (*s) {
		if (*s == c)
			p = s;
		++s;
		}
	return p;
	}

/* end LSTRING.C */
