/*
**	File: rominteg.c
**
**
**	Notes:
**		   includes the short/long integer only version of
**		   (a) _fmt which supports the sprintf functions,
**		   (b) sscanf.
**		   The floating point versions of all these
**		   routines are in romfloat.c
**
**	Copyright (c) 1984, 1985 Quality Computer Systems
**	All rights reserved.
**
**	Published by:	The Code Works
**			Box 62136
**			Santa Barbara, CA 93160
**
**	Modified: 4 Apr 85 rj
**		 11 Sep 85 jc
**		 16 Apr 93 ACZ - reduced to the funcions needed for
**				 ROM-programs in EMUFs
*/

#include "romio.h"

/* Library parameters */

#define	TRUE	1
#define	FALSE	0
#define MAXFLD	128	/* maximum field width for printf formats */
#define MAXSCAN 128	/* maximum size of string in _scanf %[] format */

/*: _fmt -- Internal routine for printf/fprintf/sprintf (integer only) */
_fmt(func, dest, format, args)
register (*func)();
register char *dest, **args;
char *format;
{
/*
 * WARNING: the static variables used in this function for compactness
 *	    prevent it from being used in an interrupt environment
 *	    where this routine can be overlaid and then reloaded.
 *	    Change all static declarations to auto or register to
 *	    eliminate this problem.
 */
	static char *sptr;
	register char *wptr;
	register int c;
	static base, pf, ljf, padchr, width, precision;
	int longflag;
	char wbuf[MAXFLD], *itob(), *ltob();

	while (c = *format++) {
	    if (c == '%') {
		wptr = wbuf;
		precision = 6;
		ljf = pf = longflag = FALSE;

		if (*format == '-') {
			++format;
			ljf = TRUE;
			}
		padchr = (*format == '0') ? '0': ' ';
		width = (isdigit(*format)) ? _atoi(&format) : 0;
		if ((c = *format++) == '.') {
			precision = _atoi(&format);
			pf = TRUE;
			c = *format++;
			}
		if (toupper(c) == 'L') {
			longflag = TRUE;
			c = *format++;
			}
		switch(toupper(c)) {

		case 'D':
			base = -10; goto val;
		case 'U':
			base = 10; goto val;
		case 'X':
			base = 16; goto val;
		case 'O':
			base = 8;
		val:
			if (longflag) {
				ltob( *(long *)args, wptr, base);
				args += 2;
				}
			else
				itob( *(int *)args++, wptr, base);
			width -= strlen(wptr);
			goto pad2;
		case 'C':
			*wptr++ = *args++;
			--width;
			goto pad;
		case 'S':
			if (!pf) precision = MAXFLD;
			sptr = *args++;
			while (*sptr && precision) {
				*wptr++ = *sptr++;
				--precision;
				--width;
			}
		pad:	*wptr = '\0';
		pad2:	wptr = wbuf;
			if (!ljf)
				while (--width >= 0)
					(*func)(padchr,dest);
			while (*wptr)
				(*func)(*wptr++,dest);
			if (ljf)
				while (--width >= 0)
					(*func)(padchr,dest);
			break;

		default:
			(*func)(c,dest);
		}
	    }
	    else (*func)(c,dest);
	}
/* end _fmt */ }

/*: sscanf -- Get formatted input from a string */
sscanf(s, fmt, args)
char *s, *fmt, *args;
	{
	char ss[MAXSCAN+1], *strchr();
	static char *p, **pargs;
	static count, c, f, include, present, quit, suppress, width,
		base, d, maxd, longflag, sign;
	static long value;

	pargs = &args;
	count = 0;
	quit = FALSE;
	c = *s++;
	while (f = *fmt++) {
		if (isspace(f)) 	/* skip white space in format */
			continue;
		if (f != '%') { 	/* must match next input string */
			while (isspace(c))
				c = *s++;
			if (f == c) {
				c = *s++; /* matches so get next char */
				continue;
				}
			else
				break;
			}
		if (*fmt == '*') {
			suppress = TRUE;
			++fmt;
			}
		else
			suppress = FALSE;
		width = (isdigit(*fmt))? _atoi(&fmt): -1;
		if (*fmt == 'h') {	/* short int modifier */
			longflag = FALSE;
			++fmt;
			}
		else if (*fmt == 'l') {	/* long integer modifier */
			longflag = TRUE;
			++fmt;
			}
		else
			longflag = FALSE;
		switch (*fmt++) {
		case '%':
			while (isspace(c))
				c = *s++;
			if (c == '%') {
				c = *s++;
				continue;
				}
			else {
				quit = TRUE;
				break;
				}
		case '[':
			if (*fmt != '^')
				include = TRUE; /* include chars in [] */
			else {
				include = FALSE;
				++fmt;
				}
			for (d = 0; d < MAXSCAN; ++d) {
				f = *fmt++;	/* get char to incl/excl */
				if (f == ']')
					break;
				else if (f == '\0') {
					quit = TRUE;
					break;
					}
				else
					ss[d] = f;
				}
			ss[d] = '\0';
			if (quit)
				break;		/* format didn't end with ] */
			while (isspace(c))
				c = *s++;
			if (c == '\0')
				break;
			if (!suppress) {
				++count;
				s = *pargs++;
				}
			while (width--) {
				present = (strchr(ss, c) != NULL);
				if ((include && !present)
				   || (!include && present))
					break;	/* doesn't meet %[s] */
				if (!suppress)
					*s++ = c;
				if ((c = *s++) == '\0')
					break;	/* end of input string */
				}
			if (!suppress)
				*s = '\0';
			break;
		case 'X':
			longflag = TRUE;
		case 'x':
			base = 16;
			maxd = 'F';
			sign = 1;
			goto getnum;
		case 'O':
			longflag = TRUE;
		case 'o':
			base = 8;
			maxd = '7';
			sign = 1;
			goto getnum;
		case 'D':
			longflag = TRUE;
		case 'd':
			base = 10;
			maxd = '9';
			while (isspace(c))
				c = *s++;
			if (c == '-') {
				sign = -1;
				c = *s++;
				--width;
				}
			else if (c == '+') {
				sign = 1;
				c = *s++;
				--width;
				}
			else
				sign = 1;
			goto getnum1;
		getnum:
			while (isspace(c))
				c = *s++;
		getnum1:
			d = toupper(c);
			if (isalnum(d) && d <= maxd && width) {
				if (!suppress)
					++count;
				}
			else {
				quit = TRUE;
				break;
				}
			value = 0;
			while (width--) {
				d = toupper(c);
				if (isdigit(d) && d <= maxd)
					d -= '0';
				else if (isalpha(d) && d <= maxd)
					d -= ('A' - 10);
				else
					break;
				/* build negative to handle max negative int */
				value = base * value - d;
				c = *s++;
				}
			if (!suppress)
				if (sign > 0)
					value = -value; /* to handle max neg */
				if (longflag)
					*(long *)*pargs++ = value;
				else
					*(int *)*pargs++ = value;
			break;
		case 'C':
		case 'c':
			if (c == '\0')
				break;
			if (suppress) {
				if (width == -1) {
					c = *s++; /* get next character */
					continue;
					}
				while (width--) {
					if ((c=*s++) == '\0')
						break;
					}
				break;
				}
			++count;
			if (width == -1) { /* get single char */
				**pargs++ = c;
				c = *s++;
				continue;
				}
			else {		/* get string */
				p = *pargs++;
				while (width--) {
					*p++ = c;
					if ((c=*s++) == '\0')
						break;
					}
				*p = '\0';
				}
			break;
		case 'S':
		case 's':
			while (isspace(c))
				c = *s++;
			if (c == '\0')
				break;
			if (!suppress) {
				++count;
				p = *pargs++;
				}
			while (width--) {
				if (!suppress)
					*p++ = c;
				c = *s++;
				if (isspace(c) || c == '\0')
					break;
				}
			if (!suppress)
				*p = '\0';
			break;
		default:		/* can't figure this out */
			quit = TRUE;
			break;
			}
		if (c == '\0'  || quit == TRUE)
			break;
		}
	return ((count == 0 && c == '\0')? EOF: count);
/* end sscanf */	}

/*: sprintf -- format a string */
sprintf(s, format, args)
char *s, *format, *args;
	{
	int _cat();			/* function _fmt will use as output */

	*s = '\0';			/* initialize the string */
	_fmt(_cat, &s, format, &args);
	}

/* end rominteg.c */
